# -*- coding: utf-8 -*-
"""
Utilities for locating samples and variants.

See also the examples at:

- http://nbviewer.ipython.org/github/alimanfoo/anhima/blob/master/examples/loc.ipynb

"""  # noqa


from __future__ import division, print_function, absolute_import


# standard library dependencies
import random


# third party dependencies
import numpy as np
import numexpr
import matplotlib.pyplot as plt
import scipy.stats


def view_sample(a, selection, all_samples=None):
    """View a single column from the array `a` corresponding to a selected
    sample.

    Parameters
    ----------

    a : array_like
        An array with 2 or more dimensions, where the second dimension
        corresponds to samples.
    selection : int or object
        A sample identifier or column index.
    all_samples : sequence, optional
        A sequence (e.g., list) of sample identifiers corresponding to the
        second dimension of `a`, used to map `selection` to a column index. If
        not given, assume `selection` is a column index.

    Returns
    -------

    b : ndarray
        An array obtained from `a` by taking the column corresponding to the
        selected sample.

    """

    # check a is an array of 2 or more dimensions
    a = np.asarray(a)
    assert a.ndim > 1

    if all_samples is None:

        # assume selection is an index
        index = selection

    else:

        # check length of samples dimension is as expected
        assert a.shape[1] == len(all_samples)

        # make sure it's a list
        all_samples = list(all_samples)

        # check selection is in all_samples
        assert selection in all_samples

        # determine indices for selected samples
        index = all_samples.index(selection)

    # view a column from the array
    b = a[:, index, ...]

    return b


def take_samples(a, selection, all_samples=None):
    """Extract columns from the array `a` corresponding to selected samples.

    Parameters
    ----------

    a : array_like
        An array with 2 or more dimensions, where the second dimension
        corresponds to samples.
    selection : sequence of ints or objects
        A sequence of sample identifiers or column indices.
    all_samples : sequence, optional
        A sequence (e.g., list) of sample identifiers corresponding to the
        second dimension of `a`, used to map `selection` to column indices. If
        not given, assume `selection` is a sequence of column indices.

    Returns
    -------

    b : ndarray
        An array obtained from `a` by taking columns corresponding to the
        selected samples.

    """

    # check a is an array of 2 or more dimensions
    a = np.asarray(a)
    assert a.ndim > 1

    if all_samples is None:

        # assume selection is column indices
        indices = selection

    else:

        # check length of samples dimension is as expected
        assert a.shape[1] == len(all_samples)

        # make sure it's a list
        all_samples = list(all_samples)

        # check selections are in all_samples
        assert all([s in all_samples for s in selection])

        # determine indices for selected samples
        indices = [all_samples.index(s) for s in selection]

    # take columns from the array
    b = np.take(a, indices, axis=1)

    return b


def query_variants(expression, variants):
    """Evaluate `expression` with respect to the given `variants`.

    Parameters
    ----------

    expression : string
        The query expression to apply. The expression will be evaluated by
        :mod:`numexpr` against the provided `variants`.
    variants : dict-like
        The variables to include in scope for the expression evaluation.

    Returns
    -------

    result : ndarray
        The result of evaluating `expression` against `variants`.

    """

    result = numexpr.evaluate(expression, local_dict=variants)

    return result


def compress_variants(a, condition):
    """Extract rows from the array `a` corresponding to a boolean `condition`.

    Parameters
    ----------

    a :  array_like
        An array to extract rows from (e.g., genotypes).
    condition : array_like, bool
        A 1-D boolean array of the same length as the first dimension of `a`.

    Returns
    -------

    b : ndarray
        An array obtained from `a` by taking rows corresponding to the
        selected variants.

    See Also
    --------

    take_variants, numpy.compress

    """

    # check dimensions and sizes
    a = np.asarray(a)
    condition = np.asarray(condition)
    assert a.ndim >= 1
    assert condition.ndim == 1
    assert a.shape[0] == condition.shape[0]

    # compress rows from the input array
    b = np.compress(condition, a, axis=0)

    return b


def take_variants(a, indices, mode='raise'):
    """Extract rows from the array `a` corresponding to `indices`.

    Parameters
    ----------

    a :  array_like
        An array to extract rows from (e.g., genotypes).
    indices : sequence of integers
        The variant indices to extract.
    mode : {'raise', 'wrap', 'clip'}, optional
        Specifies how out-of-bounds indices will behave.

    Returns
    -------

    b : ndarray
        An array obtained from `a` by taking rows corresponding to the
        selected variants.

    See Also
    --------

    compress_variants, numpy.take

    """

    # check dimensions and sizes
    a = np.asarray(a)
    assert a.ndim >= 1

    # take rows from the input array
    b = np.take(a, indices, axis=0, mode=mode)

    return b


def locate_position(pos, p):
    """Locate the index of coordinate `p` within sorted array of genomic
    positions `pos`.

    Parameters
    ----------

    pos : array_like
        A sorted 1-dimensional array of genomic positions from a single
        chromosome/contig, with no duplicates.
    p : int
        The position to locate.

    Returns
    -------

    index : int or None
        The index of `p` in `pos` if present, else None.

    See Also
    --------

    locate_positions, locate_interval, locate_intervals

    """

    # check inputs
    pos = np.asarray(pos)

    # find position
    index = np.searchsorted(pos, p)
    if index < pos.size and pos[index] == p:
        return index
    else:
        return None


def view_position(a, pos, p):
    """View a slice along the first dimension of `a` corresponding to a
    genome position.

    Parameters
    ----------

    a : array_like
        The array to extract from.
    pos : array_like
        A sorted 1-dimensional array of genomic positions from a single
        chromosome/contig, with no duplicates.
    p : int
        The position to locate.

    Returns
    -------

    b : ndarray
        A view of `a` obtained by slicing along the first dimension.

    See Also
    --------

    locate_position

    """

    # normalise inputs
    a = np.asarray(a)

    # determine region slice
    index = locate_position(pos, p)

    if index is not None:
        return a[index, ...]
    else:
        return None


def locate_interval(pos, start_position=0, stop_position=None):
    """Locate the start and stop indices within the `pos` array that include all
    positions within the `start_position` and `stop_position` range.

    Parameters
    ----------

    pos : array_like
        A sorted 1-dimensional array of genomic positions from a single
        chromosome/contig.
    start_position : int
        Start position of interval.
    stop_position : int
        Stop position of interval

    Returns
    -------

    loc : slice
        A slice object with the start and stop indices that include all
        positions within the interval.

    See Also
    --------

    locate_position, locate_positions, locate_intervals

    """

    # check inputs
    pos = np.asarray(pos)

    # locate start and stop indices
    start_index = np.searchsorted(pos, start_position)
    stop_index = np.searchsorted(pos, stop_position, side='right') \
        if stop_position is not None else None

    loc = slice(start_index, stop_index)
    return loc


def view_interval(a, pos, start_position, stop_position):
    """View a contiguous slice along the first dimension of `a`
    corresponding to a genome interval defined by `start_position` and
    `stop_position`.

    Parameters
    ----------

    a : array_like
        The array to extract from.
    pos : array_like
        A sorted 1-dimensional array of genomic positions from a single
        chromosome/contig.
    start_position : int
        Start position of interval.
    stop_position : int
        Stop position of interval

    Returns
    -------

    b : ndarray
        A view of `a` obtained by slicing along the first dimension.

    See Also
    --------

    locate_interval

    """

    # normalise inputs
    a = np.asarray(a)

    # determine region slice
    loc = locate_interval(pos, start_position, stop_position)

    return a[loc, ...]


def locate_positions(pos1, pos2):
    """Find the intersection of two sets of positions.

    Parameters
    ----------

    pos1, pos2 : array_like
        A sorted 1-dimensional array of genomic positions from a single
        chromosome/contig, with no duplicates.

    Returns
    -------

    cond1 : ndarray, bool
        An array of the same length as `pos1` where an element is True if the
        corresponding item in `pos1` is also found in `pos2`.
    cond2 : ndarray, bool
        An array of the same length as `pos2` where an element is True if the
        corresponding item in `pos2` is also found in `pos1`.

    See Also
    --------

    locate_position, locate_interval, locate_intervals

    """

    # check inputs
    pos1 = np.asarray(pos1)
    pos2 = np.asarray(pos2)

    # find intersection
    cond1 = np.in1d(pos1, pos2, assume_unique=True)
    cond2 = np.in1d(pos2, pos1, assume_unique=True)

    return cond1, cond2


def locate_intervals(pos, start_positions, stop_positions):
    """Locate items within the `pos` array that fall within any of the
    intervals given by `start_positions` and `stop_positions`.

    Parameters
    ----------

    pos : array_like
        A sorted 1-dimensional array of genomic positions from a single
        chromosome/contig.
    start_positions : array_like, int
        Start positions of intervals.
    stop_positions : array_like, int
        Stop positions of intervals

    Returns
    -------

    cond1 : ndarray, bool
        An array of the same length as `pos` where an element is True if the
        corresponding item in `pos` is also found in any of the intervals.
    cond2 : ndarray, bool
        An array of the same length as the number of intervals, where an
        element is True if the corresponding interval contains one or more
        positions in `pos`.

    See Also
    --------

    locate_position, locate_positions, locate_interval

    """

    # check inputs
    pos = np.asarray(pos)
    assert pos.ndim == 1
    start_positions = np.asarray(start_positions)
    stop_positions = np.asarray(stop_positions)
    assert start_positions.ndim == stop_positions.ndim == 1
    assert start_positions.shape[0] == stop_positions.shape[0]

    # find indices of start and stop positions in pos
    start_indices = np.searchsorted(pos, start_positions)
    stop_indices = np.searchsorted(pos, stop_positions, side='right')

    # find intervals overlapping at least one position
    cond2 = start_indices < stop_indices

    # find positions within at least one interval
    cond1 = np.zeros_like(pos, dtype=np.bool)
    for i, j in zip(start_indices[cond2], stop_indices[cond2]):
        cond1[i:j] = True

    return cond1, cond2


def plot_variant_locator(pos, step=1, ax=None, start_position=None,
                         stop_position=None, flip=False, line_args=None):
    """
    Plot lines indicating the physical genome location of variants. By
    default the top x axis is in variant index space, and the bottom x axis
    is in genome position space.

    Parameters
    ----------

    pos : array_like
        A sorted 1-dimensional array of genomic positions from a single
        chromosome/contig.
    step : int, optional
        Plot a line for every `step` variants.
    ax : axes, optional
        The axes on which to draw. If not provided, a new figure will be
        created.
    start_position : int, optional
        The start position for the region over which to work.
    stop_position : int, optional
        The stop position for the region over which to work.
    flip : bool, optional
        Flip the plot upside down.
    line_args : dict-like
        Additional keyword arguments passed through to `plt.Line2D`.

    Returns
    -------

    ax : axes
        The axes on which the plot was drawn

    """

    # set up axes
    if ax is None:
        fig = plt.figure(figsize=(7, 1))
        ax = fig.add_subplot(111)

    # determine x axis limits
    if start_position is None:
        start_position = np.min(pos)
    if stop_position is None:
        stop_position = np.max(pos)
    ax.set_xlim(start_position, stop_position)

    # plot the lines
    if line_args is None:
        line_args = dict()
    line_args.setdefault('linewidth', .5)
    n_variants = len(pos)
    for i, p in enumerate(pos[::step]):
        xfrom = p
        xto = (
            start_position +
            ((i * step / n_variants) * (stop_position-start_position))
        )
        l = plt.Line2D([xfrom, xto], [0, 1], **line_args)
        ax.add_line(l)

    # invert?
    if flip:
        ax.invert_yaxis()
        ax.xaxis.tick_top()
    else:
        ax.xaxis.tick_bottom()

    # tidy up
    ax.set_yticks([])
    for l in 'left', 'right':
        ax.spines[l].set_visible(False)

    return ax


def windowed_variant_counts(pos, window_size, start_position=None,
                            stop_position=None):
    """Count variants in non-overlapping windows over the genome.

    Parameters
    ----------

    pos : array_like
        A sorted 1-dimensional array of genomic positions from a single
        chromosome/contig.
    window_size : int
        The size in base-pairs of the windows.
    start_position : int, optional
        The start position for the region over which to work.
    stop_position : int, optional
        The stop position for the region over which to work.

    Returns
    -------

    counts : ndarray, int
        The number of variants in each window.
    bin_edges : ndarray, int
        The edge positions of each window. Note that this has length
        ``len(counts)+1``. To determine bin centers use
        ``(bin_edges[:-1] + bin_edges[1:]) / 2``. To determine bin widths use
        ``np.diff(bin_edges)``.

    See Also
    --------

    windowed_variant_counts_plot, windowed_variant_density

    """

    # determine bins
    if stop_position is None:
        stop_position = np.max(pos)
    if start_position is None:
        start_position = np.min(pos)
    bin_edges = np.append(
        np.arange(start_position, stop_position, window_size), stop_position
    )

    # make a histogram of positions
    counts, _ = np.histogram(pos, bins=bin_edges)

    return counts, bin_edges


def plot_windowed_variant_counts(pos, window_size, start_position=None,
                                 stop_position=None,
                                 ax=None, plot_kwargs=None):
    """Plot windowed variant counts.

    Parameters
    ----------

    Parameters
    ----------

    pos : array_like
        A sorted 1-dimensional array of genomic positions from a single
        chromosome/contig.
    window_size : int
        The size in base-pairs of the windows.
    start_position : int, optional
        The start position for the region over which to work.
    stop_position : int, optional
        The stop position for the region over which to work.
    ax : axes, optional
        The axes on which to draw. If not provided, a new figure will be
        created.
    plot_kwargs : dict-like
        Additional keyword arguments passed through to `plt.plot`.

    Returns
    -------

    ax : axes
        The axes on which the plot was drawn.

    See Also
    --------

    windowed_variant_counts, windowed_variant_density_plot

    """

    # set up axes
    if ax is None:
        fig = plt.figure(figsize=(7, 2))
        ax = fig.add_subplot(111)

    # count variants
    y, bin_edges = windowed_variant_counts(pos, window_size,
                                           start_position=start_position,
                                           stop_position=stop_position)

    # calculate bin centers
    bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2

    # plot data
    if plot_kwargs is None:
        plot_kwargs = dict()
    plot_kwargs.setdefault('linestyle', '-')
    plot_kwargs.setdefault('marker', None)
    ax.plot(bin_centers, y, **plot_kwargs)

    # tidy up
    ax.set_ylim(bottom=0)
    ax.set_xlabel('position')
    ax.set_ylabel('count')
    if start_position is None:
        start_position = np.min(pos)
    if stop_position is None:
        stop_position = np.max(pos)
    ax.set_xlim(start_position, stop_position)

    return ax


def windowed_variant_density(pos, window_size, start_position=None,
                             stop_position=None):
    """Calculate per-base-pair density of variants in non-overlapping windows
    over the genome.

    Parameters
    ----------

    pos : array_like
        A sorted 1-dimensional array of genomic positions from a single
        chromosome/contig.
    window_size : int
        The size in base-pairs of the windows.
    start_position : int, optional
        The start position for the region over which to work.
    stop_position : int, optional
        The stop position for the region over which to work.

    Returns
    -------

    density : ndarray, int
        The density of variants in each window.
    bin_edges : ndarray, int
        The edge positions of each window. Note that this has length
        ``len(density)+1``. To determine bin centers use
        ``(bin_edges[:-1] + bin_edges[1:]) / 2``. To determine bin widths use
        ``np.diff(bin_edges)``.

    See Also
    --------

    windowed_variant_density_plot, windowed_variant_counts

    """

    # count variants in windows
    counts, bin_edges = windowed_variant_counts(pos, window_size,
                                                start_position=start_position,
                                                stop_position=stop_position)

    bin_widths = np.diff(bin_edges)

    # convert to per-base-pair density
    density = counts / bin_widths

    return density, bin_edges


def plot_windowed_variant_density(pos, window_size, start_position=None,
                                  stop_position=None,
                                  ax=None, plot_kwargs=None):
    """Plot windowed variant density.

    Parameters
    ----------

    Parameters
    ----------

    pos : array_like
        A sorted 1-dimensional array of genomic positions from a single
        chromosome/contig.
    window_size : int
        The size in base-pairs of the windows.
    start_position : int, optional
        The start position for the region over which to work.
    stop_position : int, optional
        The stop position for the region over which to work.
    ax : axes, optional
        The axes on which to draw. If not provided, a new figure will be
        created.
    plot_kwargs : dict-like
        Additional keyword arguments passed through to `plt.plot`.

    Returns
    -------

    ax : axes
        The axes on which the plot was drawn.

    See Also
    --------

    windowed_variant_density, windowed_variant_counts_plot

    """

    # set up axes
    if ax is None:
        fig = plt.figure(figsize=(7, 2))
        ax = fig.add_subplot(111)

    # count variants
    y, bin_edges = windowed_variant_density(pos, window_size,
                                            start_position=start_position,
                                            stop_position=stop_position)

    # calculate bin centers
    bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2

    # plot data
    if plot_kwargs is None:
        plot_kwargs = dict()
    plot_kwargs.setdefault('linestyle', '-')
    plot_kwargs.setdefault('marker', None)
    ax.plot(bin_centers, y, **plot_kwargs)

    # tidy up
    ax.set_ylim(bottom=0)
    ax.set_xlabel('position')
    ax.set_ylabel('density')
    if start_position is None:
        start_position = np.min(pos)
    if stop_position is None:
        stop_position = np.max(pos)
    ax.set_xlim(start_position, stop_position)

    return ax


def windowed_statistic(pos, values, window_size,
                       start_position=None,
                       stop_position=None,
                       statistic='mean'):
    """Calculate a statistic for `values` binned in non-overlapping windows
    over the genome.

    Parameters
    ----------

    pos : array_like
        A sorted 1-dimensional array of genomic positions from a single
        chromosome/contig.
    values : array_like
        A 1-D array of the same length as `pos`.
    window_size : int
        The size in base-pairs of the windows.
    start_position : int, optional
        The start position for the region over which to work.
    stop_position : int, optional
        The stop position for the region over which to work.
    statistic : string or function
        The function to apply to values in each bin.

    Returns
    -------

    stats : ndarray
        The values of the statistic within each bin.
    bin_edges : ndarray
        The edge positions of each window. Note that this has length
        ``len(stats)+1``. To determine bin centers use
        ``(bin_edges[:-1] + bin_edges[1:]) / 2``. To determine bin widths use
        ``np.diff(bin_edges)``.

    """

    # determine bins
    if stop_position is None:
        stop_position = np.max(pos)
    if start_position is None:
        start_position = np.min(pos)
    bin_edges = np.append(
        np.arange(start_position, stop_position, window_size), stop_position
    )

    # compute binned statistic
    stats, _, _ = scipy.stats.binned_statistic(pos, values=values,
                                               statistic=statistic,
                                               bins=bin_edges)

    return stats, bin_edges


def evenly_downsample_variants(a, k):
    """Evenly downsample an array along the first dimension to length `k` (or
    as near as possible), assuming the first dimension corresponds to variants.

    Parameters
    ----------

    a : array_like
        The array to downsample.
    k : int
        The target number of variants.

    Returns
    -------

    b : array_like
        A downsampled view of `a`.

    """

    # normalise inputs
    a = np.asarray(a)

    # determine length of first dimension
    n_variants = a.shape[0]

    # determine step
    step = max(1, int(n_variants/k))

    # take slice
    b = a[::step, ...]

    return b


def randomly_downsample_variants(a, k):
    """Evenly downsample an array along the first dimension to length `k`,
    assuming the first dimension corresponds to variants.

    Parameters
    ----------

    a : array_like
        The array to downsample.
    k : int
        The k number of variants.

    Returns
    -------

    b : array_like
        A downsampled copy of `a`.

    """

    # normalise inputs
    a = np.asarray(a)

    # determine length of first dimension
    n_variants = a.shape[0]

    # sample indices
    indices = sorted(random.sample(range(n_variants), k))

    # apply selection
    b = np.take(a, indices, axis=0)

    return b



# -*- coding: utf-8 -*-
"""
This module provides some facilities for constructing and plotting trees. It
is mostly a wrapper around a very limited subset of functions from the R
`ape` package (Analyses of Phylogenetics and Evolution).

R must be installed, the `ape` R package must be installed, and the Python
package ``rpy2`` must be installed, e.g.::

    $ apt-get install r-base
    $ pip install rpy2
    $ R
    > install.packages("ape")

See also the examples at:

- http://nbviewer.ipython.org/github/alimanfoo/anhima/blob/master/examples/tree.ipynb

"""  # noqa


from __future__ import division, print_function, absolute_import


# standard library dependencies
import tempfile
import logging


# third party dependencies
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.image as mpimg


logger = logging.getLogger(__name__)
debug = logging.debug


_r_initialised = False
r = None
ro = None
grdevices = None
ape = None


def _init_r():
    """Private function to initialise R, only executed when needed.

    """

    global _r_initialised
    global r
    global ro
    global grdevices
    global ape

    if not _r_initialised:

        import rpy2.robjects as ro  # noqa
        from rpy2.robjects import r
        from rpy2.robjects.packages import importr
        import rpy2.robjects.numpy2ri as numpy2ri
        numpy2ri.activate()
        grdevices = importr('grDevices')
        ape = importr(
            'ape',
            robject_translations={
                'delta.plot': 'delta_dot_plot',
                'dist.dna': 'dist_dot_dna',
                'dist.nodes': 'dist_dot_nodes',
                'node.depth': 'node_dot_depth',
                'node.depth.edgelength': 'node_dot_depth_dot_edgelength',
                'node.height': 'node_dot_height',
                'node.height.clado': 'node_dot_height_dot_clado',
                'prop.part': 'prop_dot_part',
            }
        )

        # Define custom R functions to help with coloring tree edges by
        # population. These functions were written by Jacob Almagro-Garcia
        # <jg10@sanger.ac.uk> at the Wellcome Trust Sanger Institute.
        r("""
library(ape)


######################################################################################################################
#' Computes the number of leaves of each group that hang from each branch.
#' @param phylotree A tree of class phylo.
#' @param labelgroups A vector with the group of the tip labels (named with the labels).
#' @return A named matrix with the membership counts for each interior edge of the tree.
######################################################################################################################

computeEdgeGroupCounts <- function(phylotree, labelgroups) {

  labels <- phylotree$tip.label
  num_tips <- length(labels)
  edge_names <- unique(sort(c(phylotree$edge)))

  # This matrix will keep track of the group counts for each edge.
  edge_group_counts <- matrix(0, nrow=length(edge_names), ncol=length(unique(sort(labelgroups))))
  rownames(edge_group_counts) <- edge_names
  colnames(edge_group_counts) <- unique(labelgroups)

  # Init the leaf branches.
  sapply(1:num_tips, function(l) {
    edge_group_counts[as.character(l), as.character(labelgroups[phylotree$tip.label[l]])] <<- 1
  })

  # Sort edges by the value of the descendent
  # The first segment will contain the leaves whereas the second the branches (closer to leaves first).
  # We need to do this because leaves are numbered 1:num_tips and the branches CLOSER to the leaves
  # with higher numbers.
  edges <- phylotree$edge[order(phylotree$edge[,2]),]
  branches <- edges[num_tips:nrow(edges),]
  edges[num_tips:nrow(edges),] <- branches[order(branches[,1],decreasing=T),]
  invisible(apply(edges, 1, function(edge) {
    # Check if we are connecting a leaf.
    if(edge[2] <= num_tips) {
      e <- as.character(edge[1])
      g <- as.character(labelgroups[phylotree$tip.label[edge[2]]])
      edge_group_counts[e,g] <<- edge_group_counts[e,g] + 1
    }
    else {
      e1 <- as.character(edge[1])
      e2 <- as.character(edge[2])
      edge_group_counts[e1,] <<- edge_group_counts[e1,] + edge_group_counts[e2,]
    }
  }))
  return(edge_group_counts)
}


######################################################################################################################
#' Assigns the color of the majority group (hanging from) each branch.
#' @param phylotree A tree of class phylo.
#' @param edge_group_counts A named matrix with the group counts for each branch.
#' @param groupcolors A named vector with the color of each group.
#' @param equality_color The color to be used if there is no majority group.
#' @return A vector with the colors to be used with the tree branches.
######################################################################################################################

assignMajorityGroupColorToEdges <- function(phylotree, edge_group_counts, groupcolors, equality_color="gray") {
  edge_colors <- apply(phylotree$edge, 1, function(branch) {
    e <- as.character(branch[2])
    major_group_index <- which.max(edge_group_counts[e,])
    if(all(edge_group_counts[e,] == edge_group_counts[e,major_group_index]))
      return(equality_color)
    else
      return(groupcolors[colnames(edge_group_counts)[major_group_index]])
  })
  return(edge_colors)
}
""")  # noqa

        _r_initialised = True


def nj(dist_square, labels=None):
    """Wrapper for the `ape` ``nj`` function, which performs the
    neighbor-joining tree estimation of Saitou and Nei (1987).

    Parameters
    ----------

    dist_square : array_like, shape (`n_samples`, `n_samples`)
        A pairwise distance matrix in square form.
    labels : sequence of strings, optional
        A sequence of strings to label the tips of the tree. Must be in the
        same order as rows of the distance matrix.

    Returns
    -------

    An R object of class "phylo".

    See Also
    --------

    anhima.dist.pairwise_distance

    """

    # setup R
    _init_r()

    # normalise inputs
    dist_square = np.asarray(dist_square)
    assert dist_square.ndim == 2
    assert dist_square.shape[0] == dist_square.shape[1]

    # convert distance matrix to R
    m = ro.vectors.Matrix(dist_square)

    # assign row and column labels
    if labels:
        # map all strings to str
        labels = [str(l) for l in labels]
        s = ro.StrVector(labels)
        m.rownames = s
        m.colnames = s

    # build the tree
    tree = ape.nj(m)

    return tree


def bionj(dist_square, labels=None):
    """Wrapper for the `ape` ``bionj`` function, which performs the BIONJ
    algorithm of Gascuel (1997).

    Parameters
    ----------

    dist_square : array_like, shape (`n_samples`, `n_samples`)
        A pairwise distance matrix in square form.
    labels : sequence of strings, optional
        A sequence of strings to label the tips of the tree. Must be in the
        same order as rows of the distance matrix.

    Returns
    -------

    An R object of class "phylo".

    See Also
    --------

    anhima.dist.pairwise_distance

    """

    # setup R
    _init_r()

    # normalise inputs
    dist_square = np.asarray(dist_square)
    assert dist_square.ndim == 2
    assert dist_square.shape[0] == dist_square.shape[1]

    # convert distance matrix to R
    m = ro.vectors.Matrix(dist_square)

    # assign row and column labels
    if labels:
        # map all strings to str
        labels = [str(l) for l in labels]
        s = ro.StrVector(labels)
        m.rownames = s
        m.colnames = s

    # build the tree
    tree = ape.bionj(m)

    return tree


def plot_phylo(tree, plot_kwargs=None, add_scale_bar=None,
               filename=None, width=None, height=None, units=None, res=None,
               pointsize=None, bg=None, ax=None, imshow_kwargs=None):
    """Wrapper for the `ape` ``plot.phylo`` function, which plots phylogenetic
    trees. Plotting will use the R `png` graphics device.

    Parameters
    ----------

    tree : R object of class "phylo"
        The tree to plot.
    plot_kwargs : dict-like, optional
        A dictionary of keyword arguments that will be passed through to the
        `ape` function ``plot.phylo()``. See the documentation for the `ape`
        package for a full list of supported arguments.
    add_scale_bar : dict-like, optional
        A dictionary of keyword arguments that will be passed through to the
        `ape` function ``add.scale.bar()``. See the documentation for the
        `ape` package for a full list of supported arguments.
    filename : string, optional
        File path for the generated PNG image. If None, a temporary file will
        be used.
    width : int or float, optional
        Width of the plot in `units`.
    height : int or float, optional
        Height of the plot in `units`.
    units : {'px', 'in', 'cm', 'mm'}, optional
        The units in which 'height' and 'width' are given. Can be 'px' (pixels,
        the default), 'in' (inches), 'cm' or 'mm'.
    res : int, optional
        The nominal resolution in ppi which will be recorded in the bitmap
        file, if a positive integer.  Also used for 'units' other than the
        default, and to convert points to pixels.
    pointsize : float, optional
        The default pointsize of plotted text, interpreted as big points (
        1/72 inch) at 'res' ppi.
    bg : color, optional
        The background color.
    ax : axes, optional
        The axes on which to draw. If not provided, a new figure will be
        created.
    imshow_kwargs : dict-like
        Additional keyword arguments passed through to `imshow()`.

    Returns
    -------

    ax : axes
        The axes on which the plot was drawn.

    """

    # setup R
    _init_r()

    # setup image file
    if filename is None:
        tmp = tempfile.NamedTemporaryFile(suffix='.png')
        filename = tmp.name

    # initialise PNG device
    png_arg_names = 'width', 'height', 'units', 'res', 'pointsize', 'bg'
    png_args = dict()
    for n in png_arg_names:
        v = locals()[n]
        if v is not None:
            png_args[n] = v
    debug(filename)
    debug(png_args)
    grdevices.png(filename, **png_args)

    # plot
    if plot_kwargs is None:
        plot_kwargs = dict()
    # adapt values for certain properties
    for k in 'tip.color', 'edge.color':
        if k in plot_kwargs:
            v = plot_kwargs[k]
            if isinstance(v, (list, tuple, np.ndarray)):
                plot_kwargs[k] = ro.StrVector(v)
    debug(plot_kwargs)
    ape.plot_phylo(tree, **plot_kwargs)

    # add scale bar
    if add_scale_bar is not None:
        ape.add_scale_bar(**add_scale_bar)

    # finalise PNG device
    grdevices.dev_off()

    # read in PNG for matplotlib plotting
    png = mpimg.imread(filename)

    # set up axes for matplotlib plotting
    if ax is None:
        # try to make the figure exactly the right size for image native
        # resolution
        pxw, pxh = png.shape[:2]
        dpi = plt.rcParams['savefig.dpi']
        w, h = pxw/dpi, pxh/dpi
        fig, ax = plt.subplots(figsize=(w, h))
        # no margin
        fig.subplots_adjust(0, 0, 1, 1, hspace=0, wspace=0)
    if imshow_kwargs is None:
        imshow_kwargs = dict()
    imshow_kwargs.setdefault('aspect', 'equal')
    imshow_kwargs.setdefault('interpolation', 'none')
    ax.imshow(png, **imshow_kwargs)
    ax.set_axis_off()

    return ax


def write_tree(tree, filename=None, **kwargs):
    """
    Wrapper for the `ape` ``write.tree`` function, which writes in a file a
    tree in parenthetic format using the Newick (also known as New Hampshire)
    format.

    Parameters
    ----------

    tree : R object of class "phylo"
        The tree to be written.
    filename : string, optional
        The name of the file to write to. If ommitted, write the file to a
        string and return it.
    **kwargs : keyword arguments
        All further keyword arguments are passed through to ``write.tree``.

    Returns
    -------

    result : string
        A string if `filename` is None, otherwise no return value.

    """

    # setup R
    _init_r()

    # write the file
    if filename is None:
        kwargs['file'] = ''
    else:
        kwargs['file'] = filename
    result = ape.write_tree(tree, **kwargs)

    # handle the case where tree is written to stdout
    if filename is None:
        return result[0]


def read_tree(filename, **kwargs):
    """
    Wrapper for the `ape` ``read.tree`` function, which reads a file which
    contains one or several trees in parenthetic format known as the Newick
    or New Hampshire format.

    Parameters
    ----------

    filename : string
        Name of the file to read.
    **kwargs : keyword arguments
        All further keyword arguments are passed through to ``read.tree``.

    Returns
    -------

    tree : R object of class "phylo"
        If several trees are read in the file, the returned object is of
        class "multiPhylo", and is a list of objects of class "phylo". The name
        of each tree can be specified by tree.names, or can be read from the
        file (see details).

    """

    # setup R
    _init_r()

    kwargs['file'] = filename
    return ape.read_tree(**kwargs)


def color_edges_by_group_majority(tree, labels, groups,
                                  colors,
                                  equality_color='gray'):
    """
    Color the edges of a tree according to the majority group membership of
    the descendant tips.

    Parameters
    ----------

    tree : R object of class "phylo"
        The tree containing the edges to be colored.
    labels : sequence of strings
        The tip labels.
    groups : sequence of strings
        A sequence of strings of the same length as `labels`, where each item
        is an identifier for the group to which the corresponding tip belongs.
    colors : dict-like
        A dictionary mapping groups to colors.
    equality_color : string, optional
        The color to use in the event of a tie.

    Returns
    -------

    edge_colors : list of strings
        A list of colors for the edges of the tree, to be passed into
        :func:`plot_phylo`.

    """

    # setup R
    _init_r()

    r_groups = ro.StrVector([str(g) for g in groups])
    r_groups.names = ro.StrVector([str(l) for l in labels])
    counts = r.computeEdgeGroupCounts(tree, r_groups)

    r_colors = ro.StrVector([str(v) for v in colors.values()])
    r_colors.names = ro.StrVector([str(k) for k in colors.keys()])
    edge_colors = r.assignMajorityGroupColorToEdges(
        tree, counts, groupcolors=r_colors, equality_color=equality_color
    )

    return list(edge_colors)



# -*- coding: utf-8 -*-
"""
Utilities for calculating and plotting linkage disequilbrium.

See also the examples at:

- http://nbviewer.ipython.org/github/alimanfoo/anhima/blob/master/examples/ld.ipynb

"""  # noqa


from __future__ import division, print_function, absolute_import


# third party dependencies
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.transforms as mpl_transforms
import scipy.stats as stats
import scipy.spatial.distance as distance


# internal dependencies
import anhima.loc
from anhima.opt.ld import ld_prune_pairwise_uint8 as _ld_prune_pairwise_uint8


def pairwise_genotype_ld(gn):
    """Given a set of genotypes at biallelic variants, calculate the
    square of the correlation coefficient between all distinct pairs
    of variants.

    Parameters
    ----------

    gn : array_like
        A 2-dimensional array of shape (`n_variants`, `n_samples`) where each
        element is a genotype call coded as a single integer counting the
        number of non-reference alleles.

    Returns
    -------

    r_squared : ndarray, float
        A 2-dimensional array of squared correlation coefficients between
        each pair of variants.

    """

    # check input array
    gn = np.asarray(gn)
    assert gn.ndim == 2

    # TODO deal with missing genotypes
    return np.corrcoef(gn) ** 2


def plot_pairwise_ld(r_squared, cmap='Greys', flip=True, ax=None):
    """Make a classic triangular linkage disequilibrium plot, given an
    array of pairwise correlation coefficients between variants.

    Parameters
    ----------

    r_squared : array_like
        A square 2-dimensional array of squared correlation coefficients
        between pairs of variants.
    cmap : color map, optional
        The color map to use when plotting. Defaults to 'Greys' (0=white,
        1=black).
    flip : bool, optional
        If True, draw the triangle upside down.
    ax : axes, optional
        The axes on which to draw. If not provided, a new figure will be
        created.

    Returns
    -------

    ax : axes
        The axes on which the plot was drawn

    """

    # check inputs
    r_squared = np.asarray(r_squared)
    assert r_squared.ndim == 2
    assert r_squared.shape[0] == r_squared.shape[1]

    # setup axes
    if ax is None:
        x = plt.rcParams['figure.figsize'][0]
        fig = plt.figure(figsize=(x, x//2))
        ax = fig.add_axes((0, 0, 1, 1))

    # define transformation to rotate the colormesh
    trans = mpl_transforms.Affine2D().rotate_deg_around(0, 0, -45)
    trans = trans + ax.transData

    # plot the data as a colormesh
    ax.pcolormesh(r_squared, cmap=cmap, vmin=0, vmax=1, transform=trans)

    # cut the plot in half so we see a triangle
    ax.set_ylim(bottom=0)

    # turn the triangle upside down
    if flip:
        ax.invert_yaxis()

    # remove axis lines
    ax.set_axis_off()

    return ax


def plot_windowed_ld(gn, pos, window_size, start_position=None,
                     stop_position=None, percentiles=(5, 95), ax=None,
                     median_plot_kwargs=None,
                     percentiles_plot_kwargs=None):
    """Plot average LD within non-overlapping genome windows.

    Parameters
    ----------

    gn : array_like
        A 2-dimensional array of shape (`n_variants`, `n_samples`) where each
        element is a genotype call coded as a single integer counting the
        number of non-reference alleles.
    pos : array_like
        A 1-dimensional array of genomic positions of variants.
    window_size : int
        The size in base-pairs of the windows.
    start_position : int, optional
        The start position for the region over which to work.
    stop_position : int, optional
        The stop position for the region over which to work.
    percentiles : sequence of integers, optional
        Percentiles to plot in addition to the median.
    ax : axes, optional
        The axes on which to draw. If not provided, a new figure will be
        created.
    median_plot_kwargs : dict, optional
        Keyword arguments to pass through when plotting the median line.
    percentiles_plot_kwargs : dict, optional
        Keyword arguments to pass through when plotting the percentiles.

    Returns
    -------

    ax : axes
        The axes on which the plot was drawn.

    """

    # check input array
    gn = np.asarray(gn)
    assert gn.ndim == 2

    # set up axes
    if ax is None:
        x = plt.rcParams['figure.figsize'][0]
        fig = plt.figure(figsize=(x, x//3))
        ax = fig.add_subplot(111)

    # determine bins
    if stop_position is None:
        stop_position = np.max(pos)
    if start_position is None:
        start_position = np.min(pos)
    bin_edges = np.arange(start_position, stop_position, window_size)
    n_bins = len(bin_edges) - 1

    # initialise plotting variables
    med = np.zeros((n_bins,), dtype='f4')
    if percentiles:
        pc = np.zeros((n_bins, len(percentiles)), dtype='f4')

    # iterate over bins
    for n in range(n_bins):

        # determine bin start and stop positions
        bin_start = bin_edges[n]
        bin_stop = bin_edges[n + 1]

        # map genome positions onto variant indices
        loc = anhima.loc.locate_interval(pos, bin_start, bin_stop)

        if loc.stop - loc.start > 0:

            # view genotypes for the current region
            gw = gn[loc, :]

            # calculate pairwise LD
            r_squared = pairwise_genotype_ld(gw)

            # convert to non-redundance form
            r_squared_nonredundant = distance.squareform(r_squared,
                                                         checks=False)

            # calculate median
            med[n] = np.median(r_squared_nonredundant)

            # calculate percentiles
            if percentiles:
                for i, p in enumerate(percentiles):
                    pc[n, i] = np.percentile(r_squared_nonredundant, p)

    # determine x coordinates for plotting, as bin centers
    x = (bin_edges[1:] + bin_edges[:-1]) / 2

    # plot median
    if median_plot_kwargs is None:
        median_plot_kwargs = dict()
    median_plot_kwargs.setdefault('linestyle', '-')
    median_plot_kwargs.setdefault('color', 'k')
    median_plot_kwargs.setdefault('linewidth', 2)
    ax.plot(x, med, **median_plot_kwargs)

    # plot percentiles
    if percentiles:
        if percentiles_plot_kwargs is None:
            percentiles_plot_kwargs = dict()
        percentiles_plot_kwargs.setdefault('linestyle', '--')
        percentiles_plot_kwargs.setdefault('color', 'k')
        percentiles_plot_kwargs.setdefault('linewidth', 1)
        for i, p in enumerate(percentiles):
            ax.plot(x, pc[:, i], **percentiles_plot_kwargs)

    # tidy up
    ax.set_xlabel('position')
    ax.set_ylabel('$r^2$', rotation=0)
    ax.grid(axis='y')

    return ax


def ld_prune_pairwise(gn, window_size=100, window_step=10, max_r_squared=.2):
    """Given a set of genotypes at biallelic variants, find a subset
    of the variants which are in approximate linkage equilibrium with
    each other.

    Parameters
    ----------

    gn : array_like
        A 2-dimensional array of shape (`n_variants`, `n_samples`) where each
        element is a genotype call coded as a single integer counting the
        number of non-reference alleles.
    window_size : int, optional
        The number of variants to work with at a time.
    window_step : int, optional
        The number of variants to shift the window by.
    max_r_squared : float, optional
        The maximum value of the genotype correlation coefficient, above which
        variants will be excluded.

    Returns
    -------

    included : ndarray, bool
        A boolean array of the same length as the number of variants,
        where a True value indicates the variant at the corresponding
        index is included, and a False value indicates the corresponding
        variant is excluded.

    Notes
    -----

    The algorithm is as follows. A window of `window_size` variants is
    taken from the beginning of the genotypes array. The genotype
    correlation coefficient is calculated between each pair of
    variants in the window. The first variant in the window is
    considered, and any other variants in the window with linkage
    above `max_r_squared` with respect to the first variant is
    excluded. The next non-excluded variant in the window is then
    considered, and so on. The window then shifts along by
    `window_step` variants, and the process is repeated.

    """

    # check input array
    gn = np.asarray(gn).astype('u1')
    assert gn.ndim == 2

    # use optimised implementation
    included = _ld_prune_pairwise_uint8(gn, window_size, window_step,
                                        max_r_squared)
    return included


def pairwise_ld_decay(r_squared, pos, step=1):
    """Compile data on linkage disequilibrium, separation (in number
    of variants), and physical distance between pairs of variants.

    Parameters
    ----------

    r_squared : array_like
        A square 2-dimensional array of squared correlation coefficients
        between pairs of variants.
    pos : array_like
        A 1-dimensional array of genomic positions of variants.
    step : int, optional
        When compiling the data, advance `step` variants.

    Returns
    -------

    cor : ndarray, float
        Each element in the array is the squared genotype correlation
        coefficient between a distinct pair of variants.
    sep : ndarray, int
        Each element in the array is the separation (in number of variants)
        between a distinct pair of variants.
    dist : ndarray, int
        Each element in the array is the physical distance between a distinct
        pair of variants.

    See Also
    --------

    windowed_ld_decay

    """

    # check inputs
    r_squared = np.asarray(r_squared)
    assert r_squared.ndim == 2
    assert r_squared.shape[0] == r_squared.shape[1]

    # determine the number of variants
    n_variants = r_squared.shape[0]

    # determine pairs of variants to use
    pairs = [(i, j)
             for i in range(0, n_variants, step)
             for j in range(i+1, n_variants)]

    # initialise output arrays
    cor = np.zeros((len(pairs),), dtype=np.float)
    sep = np.zeros((len(pairs),), dtype=np.int)
    dist = np.zeros((len(pairs),), dtype=np.int)

    # iterate over pairs
    for n, (i, j) in enumerate(pairs):
        cor[n] = r_squared[i, j]
        sep[n] = j - i
        dist[n] = np.abs(pos[j] - pos[i])

    return cor, sep, dist


def windowed_ld_decay(gn, pos, window_size, step=1):
    """Compile data on linkage disequilibrium, separation (in number
    of variants), and physical distance between pairs of variants.

    Parameters
    ----------

    gn : array_like
        A 2-dimensional array of shape (`n_variants`, `n_samples`) where each
        element is a genotype call coded as a single integer counting the
        number of non-reference alleles.
    pos : array_like
        A 1-dimensional array of genomic positions of variants.
    window_size : int, optional
        The number of variants to work with at a time.
    step : int, optional
        When compiling the data within each window, advance `step` variants.

    Returns
    -------

    cor : ndarray, float
        Each element in the array is the squared genotype correlation
        coefficient between a distinct pair of variants.
    sep : ndarray, int
        Each element in the array is the separation (in number of variants)
        between a distinct pair of variants.
    dist : ndarray, int
        Each element in the array is the physical distance between a distinct
        pair of variants.

    See Also
    --------

    pairwise_ld_decay

    Notes
    -----

    Similar to :func:`pairwise_ld_decay` except that not all pairs of
    variants are sampled to speed up computation and use less memory. Variants
    are divided into non-overlapping windows of size `window_size`. Genotype LD
    is calculated for all pairs within each window.

    """

    # check input array
    gn = np.asarray(gn)
    assert gn.ndim == 2

    # determine number of variants
    n_variants = gn.shape[0]

    # initialise output variables
    all_cor = list()
    all_sep = list()
    all_dist = list()

    # iterate over non-overlapping windows of variants
    for window_start in range(0, n_variants, window_size):

        # determine extent of current window
        window_stop = min(window_start + window_size, n_variants)

        # view genotypes for the current window
        gw = gn[window_start:window_stop, :]

        # calculate LD
        r_squared = pairwise_genotype_ld(gw)

        # compile data
        cor, sep, dist = pairwise_ld_decay(r_squared, pos, step=step)
        all_cor.append(cor)
        all_sep.append(sep)
        all_dist.append(dist)

    # concatenate results from each window
    all_cor = np.concatenate(all_cor)
    all_sep = np.concatenate(all_sep)
    all_dist = np.concatenate(all_dist)

    return all_cor, all_sep, all_dist


def plot_ld_decay_by_separation(cor, sep,
                                max_separation=100,
                                percentiles=(5, 95),
                                ax=None,
                                median_plot_kwargs=None,
                                percentiles_plot_kwargs=None):
    """Plot the decay of linkage disequilibrium with separation
    between variants.

    Parameters
    ----------

    cor : array_like
        A 1-dimensional array of squared correlation coefficients between
        pairs of variants.
    sep : array_like
        A 1-dimensional array of separations (in number of variants) between
        pairs of variants.
    max_separation : int, optional
        Maximum separation to consider.
    percentiles : sequence of integers, optional
        Percentiles to plot in addition to the median.
    ax : axes, optional
        The axes on which to draw. If not provided, a new figure will be
        created.
    median_plot_kwargs : dict, optional
        Keyword arguments to pass through when plotting the median line.
    percentiles_plot_kwargs : dict, optional
        Keyword arguments to pass through when plotting the percentiles.

    Returns
    -------

    ax : axes
        The axes on which the plot was drawn.

    """

    # check inputs
    cor = np.asarray(cor)
    sep = np.asarray(sep)
    assert cor.ndim == 1
    assert sep.ndim == 1
    assert cor.shape[0] == sep.shape[0]

    # set up axes
    if ax is None:
        fig, ax = plt.subplots()

    # set up arrays for plotting
    cor_median = np.zeros((max_separation,), dtype='f4')
    if percentiles:
        cor_percentiles = np.zeros((max_separation, len(percentiles)),
                                   dtype='f4')

    # iterate over separations, compiling data
    for i in range(max_separation):

        # view correlations at the given separation
        c = cor[sep == i]

        # calculate median and percentiles
        if len(c) > 0:
            cor_median[i] = np.median(c)
            if percentiles:
                for n, p in enumerate(percentiles):
                    cor_percentiles[i, n] = np.percentile(c, p)

    # plot the median
    x = range(max_separation)
    y = cor_median
    if median_plot_kwargs is None:
        median_plot_kwargs = dict()
    median_plot_kwargs.setdefault('linestyle', '-')
    median_plot_kwargs.setdefault('color', 'k')
    median_plot_kwargs.setdefault('linewidth', 2)
    plt.plot(x, y, label='median', **median_plot_kwargs)

    # plot percentiles
    if percentiles:
        if percentiles_plot_kwargs is None:
            percentiles_plot_kwargs = dict()
        percentiles_plot_kwargs.setdefault('linestyle', '--')
        percentiles_plot_kwargs.setdefault('color', 'k')
        percentiles_plot_kwargs.setdefault('linewidth', 1)
        for n, p in enumerate(percentiles):
            y = cor_percentiles[:, n]
            plt.plot(x, y, label='%s%%' % p, **percentiles_plot_kwargs)

    # tidy up
    ax.set_xlim(left=1, right=max_separation)
    ax.set_ylim(0, 1)
    ax.set_xlabel('separation')
    ax.set_ylabel('$r^2$', rotation=0)
    ax.grid(axis='y')

    return ax


def plot_ld_decay_by_distance(cor, dist, bins,
                              percentiles=(5, 95),
                              ax=None,
                              median_plot_kwargs=None,
                              percentiles_plot_kwargs=None):
    """Plot the decay of linkage disequilibrium with physical distance
    between variants.

    Parameters
    ----------

    cor : array_like
        A 1-dimensional array of squared correlation coefficients between
        pairs of variants.
    dist : array_like
        A 1-dimensional array of physical distances between pairs of variants.
    bins : int or sequence of ints
        Number of bins or bin edges. Bins of distance to calculate LD within.
    percentiles : sequence of integers, optional
        Percentiles to plot in addition to the median.
    ax : axes, optional
        The axes on which to draw. If not provided, a new figure will be
        created.
    median_plot_kwargs : dict, optional
        Keyword arguments to pass through when plotting the median line.
    percentiles_plot_kwargs : dict, optional
        Keyword arguments to pass through when plotting the percentiles.

    Returns
    -------

    ax : axes
        The axes on which the plot was drawn.

    """

    # check inputs
    cor = np.asarray(cor)
    dist = np.asarray(dist)
    assert cor.ndim == 1
    assert dist.ndim == 1
    assert cor.shape[0] == dist.shape[0]

    # set up axes
    if ax is None:
        fig, ax = plt.subplots()

    # calculate the median of correlation values within bins
    y, bin_edges, _ = stats.binned_statistic(dist, values=cor, bins=bins,
                                             statistic=np.median)

    # determine x axis variable as bin centers
    x = (bin_edges[:-1] + bin_edges[1:]) / 2

    # plot median
    if median_plot_kwargs is None:
        median_plot_kwargs = dict()
    median_plot_kwargs.setdefault('linestyle', '-')
    median_plot_kwargs.setdefault('color', 'k')
    median_plot_kwargs.setdefault('linewidth', 2)
    ax.plot(x, y, label='median', **median_plot_kwargs)

    # calculate and plot percentiles
    if percentiles:
        if percentiles_plot_kwargs is None:
            percentiles_plot_kwargs = dict()
        percentiles_plot_kwargs.setdefault('linestyle', '--')
        percentiles_plot_kwargs.setdefault('color', 'k')
        percentiles_plot_kwargs.setdefault('linewidth', 1)
        for p in percentiles:
            y, bin_edges, _ = stats.binned_statistic(
                dist,
                values=cor,
                bins=bins,
                statistic=lambda v: np.percentile(v, p)
            )
            ax.plot(x, y, label='%s%%' % p, **percentiles_plot_kwargs)

    # tidy up
    ax.set_xlim(np.min(x), np.max(x))
    ax.set_ylim(0, 1)
    ax.set_xlabel('distance')
    ax.set_ylabel('$r^2$', rotation=0)
    ax.grid(axis='y')

    return ax



# cython: profile=False
# cython: boundscheck=False
# cython: wraparound=False


from __future__ import division
import numpy as np
cimport numpy as np
from libc.math cimport sqrt


cdef inline double corrcoef_uint8(np.uint8_t[:] x, np.uint8_t[:] y):
    cdef int n = x.size
    cdef int i
    cdef int xtot = 0
    cdef int ytot = 0
    for i in range(n):
        xtot += x[i]
        ytot += y[i]
    cdef double ex = xtot/n
    cdef double ey = ytot/n
    cdef double num = 0
    cdef double xsqsumdev = 0
    cdef double ysqsumdev = 0
    for i in range(n):
        num += (x[i] - ex) * (y[i] - ey)
        xsqsumdev += (x[i] - ex)**2
        ysqsumdev += (y[i] - ey)**2
    cdef double den = sqrt(xsqsumdev) * sqrt(ysqsumdev)
    return num/den


def ld_prune_pairwise_uint8(np.uint8_t[:, :] gn,
                            int window_size=100,
                            int window_step=10,
                            float max_r_squared=.2):
    """Optimised function for pruning variants in approximate LD."""

    cdef int window_start, window_stop, n_variants, i, j
    cdef np.uint8_t[:] included, x, y
    cdef double[:, :] r_squared

    # set up output array
    n_variants = gn.shape[0]
    included = np.ones((n_variants,), dtype='u1')

    # outer loop - iterate over windows
    for window_start in range(0, n_variants, window_step):

        # determine extent of the current window
        window_stop = min(window_start + window_size, n_variants)

        # calculate pairwise genotype correlation
        if window_start == 0:
            # initialise correlation coefficients
            r_squared = np.corrcoef(gn[window_start:window_stop, :])**2
        else:
            # move up data from previous window
            r_squared[:window_size-window_step, :window_size-window_step] = \
                r_squared[window_step:, window_step:]

        # inner loop - iterate over variants within the window
        for i in range(window_stop - window_start):

            # check to see if the variant was previously excluded
            if included[window_start + i]:

                # look for linkage with other variants in window
                for j in range(i + 1, window_stop - window_start):

                    # check to see if the variant was previously excluded
                    if included[window_start + j]:

                        # do we need to calculate r_squared?
                        if j >= (window_size - window_step):
                            x = gn[window_start + i, :]
                            y = gn[window_start + j, :]
                            r_squared[i, j] = corrcoef_uint8(x, y)**2

                        if r_squared[i, j] > max_r_squared:
                            # threshold exceeded, exclude the variant
                            included[window_start + j] = 0

    return np.asarray(included).astype('b1')




# -*- coding: utf-8 -*-
"""
Input/output utilities.

"""


from __future__ import division, print_function, absolute_import


# standard library dependencies
from anhima.compat import string_types, zip


# third party dependencies
import numpy as np


def save_tped(path, genotypes, ref, alt, pos,
              chromosome='0',
              identifier=None,
              genetic_distance=None):
    """Write biallelic diploid genotype data to a file using the Plink
    transposed format (TPED).

    Parameters
    ----------

    path : string or file-like
        Path of file to write, or file-like object to write to.
    genotypes : array_like, int
        An array of shape (n_variants, n_samples, 2) where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele, etc.).
    ref : array_like, string
        A 1-dimensional array of single character strings encoding the
        reference nucleotide.
    alt : array_like, string
        A 1-dimensional array of single character strings encoding the
        alternate nucleotide.
    pos : array_like, int
        A 1-dimensional array of genomic positions.
    chromosome : string or array_like, string, optional
        Either a single string (if all variants are from the same
        chromosome/contig) or an array of strings with the chromosome of each
        variant.
    identifier : array_like, string, optional
        An array of SNP identifiers. If not provided, identifiers will be
        created based on the variant position, e.g., 'snp100042',
        'snp100081', etc.
    genetic_distance : array_like, float
        An array of genetic distances. If not provided, a zero value ('0') will
        be written for all variants.

    """

    # check genotypes
    genotypes = np.asarray(genotypes)
    assert genotypes.ndim == 3
    assert genotypes.shape[2] == 2, 'genotypes must be diploid'
    assert np.amax(genotypes) < 2, 'genotypes must be biallelic'
    n_variants = genotypes.shape[0]

    # check ref
    ref = np.asarray(ref)
    assert ref.ndim == 1
    assert ref.shape[0] == n_variants

    # check alt
    alt = np.asarray(alt)
    assert alt.ndim == 1
    assert alt.shape[0] == n_variants

    # check pos
    pos = np.asarray(pos)
    assert pos.ndim == 1
    assert pos.shape[0] == n_variants

    # check chromosome
    if isinstance(chromosome, string_types):
        chromosome = np.array([chromosome] * n_variants)
    else:
        chromosome = np.asarray(chromosome)
        assert chromosome.ndim == 1
        assert chromosome.shape[0] == n_variants

    # check identifier
    if identifier is None:
        identifier = np.array(['snp%s' % i for i in pos])
    else:
        identifier = np.asarray(identifier)
        assert identifier.shape[0] == n_variants

    # check genetic distance
    if genetic_distance is None:
        genetic_distance = np.zeros((n_variants,))
    else:
        genetic_distance = np.asarray(genetic_distance)
        assert genetic_distance.shape[0] == n_variants

    # setup output file
    tped_needs_closing = False
    if isinstance(path, string_types):
        tped_file = open(path, 'w')
        tped_needs_closing = True
    else:
        # assume file-like
        tped_file = path

    try:
        for row_data in zip(genotypes, ref, alt, pos, chromosome, identifier,
                            genetic_distance):
            out_string = _get_tped_row(*row_data)
            tped_file.write(out_string + '\n')

    finally:
        if tped_needs_closing:
            tped_file.close()


def _convert_gts_to_strings(genotypes, ref, alt):

    lu = {-1: '0', 0: ref, 1: alt}
    return [lu[a] + ' ' + lu[b] for a, b in genotypes]


def _get_tped_row(gt_data, reference, alternate, position, contig, iden,
                  genetic_dist):

    str_gts = _convert_gts_to_strings(gt_data, reference, alternate)
    return "\t".join([contig,
                      iden,
                      str(genetic_dist),
                      str(position)] + str_gts)



# -*- coding: utf-8 -*-
"""
Genetic distance calculations.

See also the examples at:

- http://nbviewer.ipython.org/github/alimanfoo/anhima/blob/master/examples/dist.ipynb

"""  # noqa


from __future__ import division, print_function, absolute_import


# third party dependencies
import numpy as np
import scipy.spatial.distance
import matplotlib.pyplot as plt


def pairwise_distance(gn, metric='euclidean'):
    """Compute pairwise distance between samples.

    Parameters
    ----------

    gn : array_like
        A 2-dimensional array of shape (n_variants, n_samples) where each
        element is a genotype call coded as a single integer counting the
        number of non-reference alleles.
    metric : string or function, optional
        The distance metric to use. See documentation for the function
        :func:`scipy.spatial.distance.pdist` for a list of supported
        distance metrics.

    Returns
    -------

    dist : ndarray, float
        The distance matrix in compact form.
    dist_square : ndarray, float
        The distance matrix in square form.

    """

    # normalise inputs
    gn = np.asarray(gn)
    assert gn.ndim == 2

    # transpose genotypes as pdist expects (m, n) for m observations in an
    # n-dimensional space
    x = gn.T

    # compute the distance matrix
    dist = scipy.spatial.distance.pdist(x, metric=metric)

    # convert to square form for easy plotting
    dist_square = scipy.spatial.distance.squareform(dist)

    return dist, dist_square


def plot_pairwise_distance(dist_square, labels=None,
                           colorbar=True, ax=None,
                           vmin=None, vmax=None, cmap='jet',
                           imshow_kwargs=None):
    """Plot pairwise distances.

    Parameters
    ----------

    dist_square : array_like
        The distance matrix in square form.
    labels : sequence of strings, optional
        Sample labels for the axes.
    colorbar : bool, optional
        If True, add a colorbar to the current figure.
    ax : axes, optional
        The axes on which to draw. If not provided, a new figure will be
        created.
    vmin : float, optional
        The minimum distance value for normalisation.
    vmax : float, optional
        The maximum distance value for normalisation.
    cmap : string, optional
        The color map for the image.
    imshow_kwargs : dict-like, optional
        Additional keyword arguments passed through to `plt.imshow`.

    Returns
    -------

    ax : axes
        The axes on which the plot was drawn

    """

    # normalise inputs
    dist_square = np.asarray(dist_square)
    assert dist_square.ndim == 2
    assert dist_square.shape[0] == dist_square.shape[1]

    # set up axes
    if ax is None:
        # make a square figure
        x = plt.rcParams['figure.figsize'][0]
        fig, ax = plt.subplots(figsize=(x, x))

    # set up normalisation
    dist = scipy.spatial.distance.squareform(dist_square)
    if vmin is None:
        vmin = np.min(dist)
    if vmax is None:
        vmax = np.max(dist)

    # plot as image
    if imshow_kwargs is None:
        imshow_kwargs = dict()
    im = ax.imshow(dist_square, interpolation='none', cmap=cmap,
                   vmin=vmin, vmax=vmax, **imshow_kwargs)

    # tidy up
    if labels:
        ax.set_xticks(range(len(labels)))
        ax.set_yticks(range(len(labels)))
        ax.set_xticklabels(labels, rotation=90)
        ax.set_yticklabels(labels, rotation=0)
    if colorbar:
        plt.gcf().colorbar(im)

    return ax



# -*- coding: utf-8 -*-
"""
Utility functions for working with data stored in HDF5 files.

**HDF5 file convention for variant call sets**

Note that this module assumes that data for a variant call set has been
organised in an HDF5 file following a particular convention. Briefly,
the convention is as follows.

An HDF5 file may contain one or more call sets. Each call set is stored
within a separate group. A call set may be stored within the root group.

Within each call set, data are grouped by chromosome.

Within each chromosome group, there are two subgroups,
named `variants` and `calldata`.

The `variants` group contains one or more datasets holding data on the
variants in the call set. The first dimension of all `variants` datasets
must have the same length, being the number of variants on the chromosome.

The `variants` group must contain a `POS` dataset holding the genome
positions of the variants.

The `calldata` group contains one or more datasets holding data relating to
genotype calls. The first dimension of all `calldata` datasets must have
the same length, being the number of variants on the chromosome. The second
dimension of all `calldata` datasets must have the same length,
being the number of samples in the cohort.

In addition to the above, a `samples` dataset may be stored within the
callset, providing a list of labels or identifiers for the samples in the
cohort. This `samples` dataset may be stored as a child of the callset group
and/or as a child of the chromosome groups.

So, for example, an HDF5 file containing a SNP call set for a cohort of
*Anopheles gambiae* samples with chromosomes (2R, 2L, 3R, 3L, X)
might be organised as follows::

    / [callset group]
    /samples [dataset, shape (n_samples,), dtype string]
    /2L [chromosome group]
    /2L/variants [variants group]
    /2L/variants/POS [dataset, shape (n_variants,), dtype int32]
    /2L/variants/REF [dataset, shape (n_variants,), dtype S1]
    /2L/variants/ALT [dataset, shape (n_variants, 3), dtype S1]
    /2L/variants/MQ [dataset, shape (n_variants,), dtype f4]
    /2L/variants/...
    /2L/calldata [calldata group]
    /2L/calldata/genotype [dataset, shape (n_variants, n_samples, ploidy), dtype int8]
    /2L/calldata/DP [dataset, shape (n_variants, n_samples), dtype=int32]
    /2L/calldata/...
    /3L/variants/...
    /3L/calldata/...
    /...

"""  # noqa


from __future__ import division, print_function, absolute_import


# standard library dependencies
import itertools
from anhima.compat import string_types, range, force_bytes


# third party dependencies
import numpy as np
import h5py
import numexpr


# internal dependencies
import anhima.loc


def load_region(callset, chrom, start_position=0, stop_position=None,
                variants_fields=None,
                calldata_fields=None,
                variants_query=None,
                samples=None):
    """Load data into memory from `callset` for the given region.

    Parameters
    ----------

    callset : HDF5 file or group
        A file or group containing a variant call set.
    chrom : string
        The chromosome to extract data for.
    start_position : int, optional
        The start position for the region to extract data for.
    stop_position : int, optional
        The stop position for the region to extract data for.
    variants_fields : sequence of strings, optional
        Names of the variants datasets to extract.
    calldata_fields : sequence of strings, optional
        Names of the calldata datasets to extract.
    variants_query : string, optional
        A query to filter variants. Note that this query is applied
        after data for the region has been loaded, so any fields
        referenced in this query need to be included in `variants_fields`.
    samples : sequence of strings, optional
        Selected samples to extract.

    Returns
    -------

    variants : dict
        A dictionary mapping dataset identifiers to ndarrays.
    calldata : dict
        A dictionary mapping dataset identifiers to ndarrays.

    """

    # obtain chromosome group
    grp_chrom = callset[chrom]

    # setup output variables
    variants = dict()
    calldata = dict()

    # obtain variant positions
    pos = grp_chrom['variants']['POS']

    # select samples needs list of all samples, check one is stored in the
    # callset and fail early if not
    all_samples = None
    if samples is not None:
        # find all samples
        if 'samples' in callset.keys():
            all_samples = list(callset['samples'])
        elif 'samples' in grp_chrom.keys():
            all_samples = list(grp_chrom['samples'])
        else:
            raise Exception('list of all samples not found in callset')

    # locate region
    loc = anhima.loc.locate_interval(pos, start_position, stop_position)

    # extract variants data
    if variants_fields:
        if isinstance(variants_fields, string_types):
            variants_fields = [variants_fields]
        for f in variants_fields:
            variants[f] = grp_chrom['variants'][f][loc, ...]

    # extract calldata
    if calldata_fields:
        if isinstance(calldata_fields, string_types):
            calldata_fields = [calldata_fields]
        for f in calldata_fields:
            calldata[f] = grp_chrom['calldata'][f][loc, ...]

    # select variants
    if variants_query is not None:
        condition = numexpr.evaluate(variants_query, local_dict=variants)
        for f in variants:
            variants[f] = np.compress(condition, variants[f], axis=0)
        for f in calldata:
            calldata[f] = np.compress(condition, calldata[f], axis=0)

    # select samples
    if samples is not None:
        # TODO check dtype of all_samples
        samples = force_bytes(samples)
        sample_indices = [all_samples.index(s) for s in samples]
        for f in calldata:
            calldata[f] = np.take(calldata[f], sample_indices, axis=1)

    return variants, calldata


def take2d_pointsel(dataset, row_indices=None, col_indices=None,
                    block_size=1000):
    """
    Load selected rows and optionally columns from an HDF5 dataset with 2 or
    more dimensions, using HDF5 point selections.

    Parameters
    ----------

    dataset : HDF5 dataset
        The dataset to load data from.
    row_indices : sequence of ints, optional
        The indices of the selected rows. If not provided, all rows will be
        returned.
    col_indices : sequence of ints, optional
        The indices of the selected columns. If not provided, all columns
        will be returned.
    block_size : int, optional
        The size (in number of points) of the block of data to load and
        process at a time.

    Returns
    -------

    out : ndarray
        An array containing the selected rows and columns.

    See Also
    --------

    anhima.util.take2d

    Notes
    -----

    This function is similar to :func:`anhima.util.take2d` but uses an HDF5
    point selection under the hood. Performance characteristics will be
    different and may be much better or much worse, depending on the size,
    shape and configuration of the dataset, and depending on the number of
    points to be selected.

    """

    n_rows_in = dataset.shape[0]
    if row_indices:
        row_indices = sorted(row_indices)
        n_rows_out = len(row_indices)
    else:
        # select all rows
        row_indices = range(n_rows_in)
        n_rows_out = n_rows_in

    n_cols_in = dataset.shape[1]
    if col_indices:
        # select all columns
        col_indices = sorted(col_indices)
        n_cols_out = len(col_indices)
    else:
        col_indices = range(n_cols_in)
        n_cols_out = n_cols_in

    n_items_out = n_rows_out * n_cols_out

    # initialise output array
    out = np.empty((n_items_out,), dtype=dataset.dtype)

    # convert indices into coordinates
    coords = itertools.product(row_indices, col_indices)

    # set up selection
    sel = h5py._hl.selections.PointSelection(dataset.shape)
    typ = h5py.h5t.py_create(dataset.dtype)

    # process blocks at a time
    for block_start in range(0, n_items_out, block_size):

        # materialise a block of coordinates
        selection = np.asarray(list(itertools.islice(coords, block_size)))

        # set selection
        sel.set(selection)

        # read data
        block_stop = block_start + len(selection)
        space = h5py.h5s.create_simple(sel.mshape)
        dataset.id.read(space,
                        sel._id,
                        out[block_start:block_stop],
                        typ)

    # reshape output array
    out = out.reshape(n_rows_out, n_cols_out)

    return out


def save_tped(path, callset, chrom,
              start_position=0,
              stop_position=None,
              samples=None):

    """Save genotype data from an HDF5 callset to a Plink transposed format
    file (TPED).

    Parameters
    ----------

    path : string or file-like
        Path of file to write, or file-like object to write to.
    callset : HDF5 file or group
        A file or group containing a variant call set.
    chrom : string
        The chromosome to extract data for.
    start_position : int, optional
        The start position for the region to extract data for.
    stop_position : int, optional
        The stop position for the region to extract data for.
    samples : sequence of strings, optional
        Selection of samples to extract genotypes for, defaults to all samples.

    Notes
    -----

    Note that the current implementation loads all data from the requested
    region into memory before writing out to TPED, so may not be applicable
    to very large datasets.

    """

    variants, calldata = load_region(callset,
                                     chrom,
                                     start_position,
                                     stop_position,
                                     variants_fields=['POS', 'REF', 'ALT'],
                                     calldata_fields=['genotype'])

    # determine samples that we will use
    if samples is None:
        genotypes = calldata['genotype']
    else:
        samples = force_bytes(samples)
        h5_samples = callset[chrom]['samples'][:].tolist()
        genotypes = np.take(
            calldata['genotype'],
            [h5_samples.index(s) for s in samples],
            axis=1)

    ref = variants['REF']

    alt = variants['ALT']
    if alt.ndim > 1:
        alt = alt[:, 0]

    pos = variants['POS']

    anhima.io.save_tped(path,
                        genotypes=genotypes,
                        ref=ref,
                        alt=alt,
                        pos=pos,
                        chromosome=chrom,
                        identifier=None,
                        genetic_distance=None)



# -*- coding: utf-8 -*-
from __future__ import division, print_function, absolute_import


import unittest
import os
import anhima.h5
import h5py
import logging


logger = logging.getLogger(__name__)
debug = logger.debug


class TestLoadRegion(unittest.TestCase):

    def setUp(self):

        ff = os.path.join('fixture', 'test32.h5')
        self.assertTrue(os.path.isfile(ff))

        self.h5_file = h5py.File(ff)
        self.samples = None
        self.start = 300000
        self.stop = 400000

    def test_load_nosamples(self):

        self.var, self.gt = anhima.h5.load_region(
            self.h5_file, '2L', self.start, self.stop,
            variants_fields=['POS', 'QD', 'DP', 'AC', 'AF'],
            calldata_fields=['genotype', 'AD', 'DP', 'GQ'],
            samples=self.samples)

        self.assertTupleEqual(self.var['POS'].shape, self.var['QD'].shape)

    def test_load_samples(self):

        self.samples = ['AC0090-C', 'AC0091-C', 'AC0092-C']
        self.var, self.gt = anhima.h5.load_region(
            self.h5_file, '2L', self.start, self.stop,
            variants_fields=['POS', 'QD', 'DP', 'AC', 'AF'],
            calldata_fields=['genotype', 'AD', 'DP', 'GQ'],
            samples=self.samples)

        self.assertTupleEqual(self.var['POS'].shape, self.var['QD'].shape)
        self.assertEqual(self.var['POS'].size, 11)

        self.assertTupleEqual(self.gt['genotype'].shape, (11, 3, 2))



from ast import literal_eval
from distutils.core import setup
from distutils.extension import Extension


try:
    from Cython.Build import cythonize
except ImportError:
    cythonize = None


if cythonize:
    ext_modules = cythonize([Extension('anhima.opt.ld',
                                       ['anhima/opt/ld.pyx'])])
else:
    ext_modules = [Extension('anhima.opt.ld', ['anhima/opt/ld.c'])]


def get_version(source='anhima/__init__.py'):
    with open(source) as f:
        for line in f:
            if line.startswith('__version__'):
                return literal_eval(line.split('=')[-1].lstrip())
    raise ValueError("__version__ not found")


setup(
    name='anhima',
    version=get_version(),
    author='Alistair Miles',
    author_email='alimanfoo@googlemail.com',
    package_dir={'': '.'},
    packages=['anhima', 'anhima.opt'],
    ext_modules=ext_modules,
    url='https://github.com/alimanfoo/anhima',
    license='MIT License',
    description='Exploration and Analysis of genetic variation data.',
    long_description=open('README.md').read(),
    classifiers=['Intended Audience :: Developers',
                 'License :: OSI Approved :: MIT License',
                 'Programming Language :: Python',
                 'Topic :: Software Development :: Libraries :: Python Modules'
                 ]
)



# -*- coding: utf-8 -*-
"""
Extremely naive simulation functions to generate genotype data for
illustration of other features in the ``anhima`` package.

"""


from __future__ import division, print_function, absolute_import


# python standard library dependencies
import random
from anhima.compat import range


# third party dependencies
import numpy as np
import scipy


def simulate_biallelic_genotypes(n_variants, n_samples, af_dist,
                                 p_missing=.1,
                                 ploidy=2):
    """Simulate genotypes at biallelic variants for a population in
    Hardy-Weinberg equilibrium

    Parameters
    ----------

    n_variants : int
        The number of variants.
    n_samples : int
        The number of samples.
    af_dist : frozen continuous random variable
        The distribution of allele frequencies.
    p_missing : float, optional
        The fraction of missing genotype calls.
    ploidy : int, optional
        The sample ploidy.

    Returns
    -------

    genotypes : ndarray, int8
        An array of shape (n_variants, n_samples, ploidy) where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = alternate allele).

    """

    # initialise output array
    genotypes = np.empty((n_variants, n_samples, ploidy), dtype='i1')

    # generate allele frequencies under the given distribution
    af = af_dist.rvs(n_variants)

    # freeze binomial distribution to model missingness
    miss_dist = scipy.stats.binom(p=p_missing, n=n_samples)

    # iterate over variants
    for i, p in zip(range(n_variants), af):

        # randomly generate alleles under the given allele frequency
        # ensure p is valid probability
        p = min(p, 1)
        alleles = scipy.stats.bernoulli.rvs(p, size=n_samples*ploidy)

        # reshape alleles as genotypes under the given ploidy
        genotypes[i] = alleles.reshape(n_samples, ploidy)

        # simulate some missingness
        n_missing = miss_dist.rvs()
        missing_indices = random.sample(range(n_samples),
                                        n_missing)
        genotypes[i, missing_indices] = (-1,) * ploidy

    return genotypes


def simulate_genotypes_with_ld(n_variants, n_samples, correlation=0.2):
    """A very simple function to simulate a set of genotypes, where
    variants are in some degree of linkage disequilibrium with their
    neighbours.

    Parameters
    ----------

    n_variants : int
        The number of variants to simulate data for.
    n_samples : int
        The number of individuals to simulate data for.
    correlation : float, optional
        The fraction of samples to copy genotypes between neighbouring
        variants.

    Returns
    -------

    gn : ndarray, int8
        A 2-dimensional array of shape (n_variants, n_samples) where each
        element is a genotype call coded as a single integer counting the
        number of non-reference alleles.

    """

    # initialise an array of random genotypes
    gn = np.random.randint(size=(n_variants, n_samples), low=0, high=3)
    gn = gn.astype('i1')

    # determine the number of samples to copy genotypes for
    n_copy = int(correlation * n_samples)

    # introduce linkage disequilibrium by copying genotypes from one sample to
    # the next
    for i in range(1, n_variants):

        # randomly pick the samples to copy from
        sample_indices = random.sample(range(n_samples), n_copy)

        # view genotypes from the previous variant for the selected samples
        c = gn[i-1, sample_indices]

        # randomly choose whether to invert the correlation
        inv = random.randint(0, 1)
        if inv:
            c = 2-c

        # copy across genotypes
        gn[i, sample_indices] = c

    return gn


def simulate_relatedness(genotypes, relatedness=.5, n_iter=1000, copy=True):
    """
    Simulate relatedness by randomly copying genotypes between individuals.

    Parameters
    ----------

    genotypes : array_like
        An array of shape (n_variants, n_samples, ploidy) where each
        element of the array is an integer corresponding to an allele index
        (-1 = missing, 0 = reference allele, 1 = first alternate allele,
        2 = second alternate allele, etc.).
    relatedness : float, optional
        Fraction of variants to copy genotypes for.
    n_iter : int, optional
        Number of times to randomly copy genotypes between individuals.
    copy : bool, optional
        If False, modify `genotypes` in place.

    Returns
    -------

    genotypes : ndarray, shape (n_variants, n_samples, ploidy)
        The input genotype array but with relatedness simulated.

    """

    # check genotypes array
    genotypes = np.asarray(genotypes)
    assert genotypes.ndim >= 2
    n_variants = genotypes.shape[0]
    n_samples = genotypes.shape[1]

    # copy input array
    if copy:
        genotypes = genotypes.copy()
    else:
        # modify in place
        pass

    # determine the number of variants to copy genotypes for
    n_copy = int(relatedness * n_variants)

    # iteratively introduce relatedness
    for i in range(n_iter):

        # randomly choose donor and recipient
        donor_index = random.randint(0, n_samples-1)
        donor = genotypes[:, donor_index]
        recip_index = random.randint(0, n_samples-1)
        recip = genotypes[:, recip_index]

        # randomly pick a set of variants to copy
        variant_indices = random.sample(range(n_variants), n_copy)

        # copy across genotypes
        recip[variant_indices] = donor[variant_indices]

    return genotypes



# -*- coding: utf-8 -*-
from __future__ import division, print_function, absolute_import


import unittest
import numpy as np
from anhima.ped import diploid_mendelian_error_biallelic, \
    diploid_mendelian_error_multiallelic


class TestMendelianErrorBiallelic(unittest.TestCase):

    def setUp(self):

        # give progeny 4 genotypes, 00, 01, 11, --
        self.progeny = np.array([[0, 0], [0, 1], [1, 1], [-1, -1]])\
            .reshape(1, -1, 2)

        # build parents with combinations of genotypes
        self.href_parent = np.zeros((1, 1, 2), dtype='int')
        self.halt_parent = np.ones((1, 1, 2), dtype='int')
        self.het_parent = np.array([0, 1], dtype='int').reshape((1, 1, 2))

    def test_homref_homref(self):
        parents = np.hstack([self.href_parent, self.href_parent])
        non_mendelian = diploid_mendelian_error_biallelic(
            parental_genotypes=parents,
            progeny_genotypes=self.progeny
        )
        self.assertTrue(np.array_equal(non_mendelian[0],
                                       [0, 1, 2, 0]))

    def test_homref_het(self):
        parents = np.hstack([self.href_parent, self.het_parent])
        non_mendelian = diploid_mendelian_error_biallelic(
            parental_genotypes=parents,
            progeny_genotypes=self.progeny
        )
        self.assertTrue(np.array_equal(non_mendelian[0], [0, 0, 1, 0]))

    def test_homref_homalt(self):
        parents = np.hstack([self.halt_parent, self.href_parent])
        non_mendelian = diploid_mendelian_error_biallelic(
            parental_genotypes=parents,
            progeny_genotypes=self.progeny,
        )
        self.assertTrue(np.array_equal(non_mendelian[0], [1, 0, 1, 0]))

    def test_homalt_het(self):
        parents = np.hstack([self.halt_parent, self.het_parent])
        non_mendelian = diploid_mendelian_error_biallelic(
            parental_genotypes=parents,
            progeny_genotypes=self.progeny,
        )
        self.assertTrue(np.array_equal(non_mendelian[0], [1, 0, 0, 0]))

    def test_homalt_homalt(self):
        parents = np.hstack([self.halt_parent, self.halt_parent])
        non_mendelian = diploid_mendelian_error_biallelic(
            parental_genotypes=parents,
            progeny_genotypes=self.progeny
        )
        self.assertTrue(np.array_equal(non_mendelian[0], [2, 1, 0, 0]))

    def test_multiple_variants(self):
        parents = np.vstack([np.hstack([self.href_parent, self.href_parent]),
                             np.hstack([self.halt_parent, self.halt_parent]),
                             np.hstack([self.het_parent, self.halt_parent]),
                             np.hstack([self.het_parent, self.het_parent])])

        multiv_progeny = np.repeat(self.progeny, 4, axis=0)

        non_mendelian = diploid_mendelian_error_biallelic(
            parental_genotypes=parents,
            progeny_genotypes=multiv_progeny
        )

        expected_result = np.array([[0, 1, 2, 0],
                                    [2, 1, 0, 0],
                                    [1, 0, 0, 0],
                                    [0, 0, 0, 0]])
        self.assertTrue(np.array_equal(parents.shape, (4, 2, 2)))
        self.assertTrue(np.array_equal(non_mendelian.shape,
                                       expected_result.shape))

        self.assertTrue(np.array_equal(non_mendelian, expected_result))


class TestMendelianErrorMultiallelic(unittest.TestCase):

    def _test(self, genotypes, expect):
        actual = diploid_mendelian_error_multiallelic(genotypes[:, 0:2],
                                                      genotypes[:, 2:], 4)
        assert np.array_equal(expect, actual)

        # swap parents, should have no affect
        actual = diploid_mendelian_error_multiallelic(genotypes[:, [1, 0]],
                                                      genotypes[:, 2:], 4)
        assert np.array_equal(expect, actual)

        # swap alleles, should have no effect
        actual = diploid_mendelian_error_multiallelic(genotypes[:, 0:2, ::-1],
                                                      genotypes[:, 2:, ::-1],
                                                      4)
        assert np.array_equal(expect, actual)

    def test_consistent(self):
        genotypes = np.array([
            # aa x aa -> aa
            [[0, 0], [0, 0], [0, 0], [-1, -1], [-1, -1], [-1, -1]],
            [[1, 1], [1, 1], [1, 1], [-1, -1], [-1, -1], [-1, -1]],
            [[2, 2], [2, 2], [2, 2], [-1, -1], [-1, -1], [-1, -1]],
            # aa x ab -> aa or ab
            [[0, 0], [0, 1], [0, 0], [0, 1], [-1, -1], [-1, -1]],
            [[0, 0], [0, 2], [0, 0], [0, 2], [-1, -1], [-1, -1]],
            [[1, 1], [0, 1], [1, 1], [0, 1], [-1, -1], [-1, -1]],
            # aa x bb -> ab
            [[0, 0], [1, 1], [0, 1], [-1, -1], [-1, -1], [-1, -1]],
            [[0, 0], [2, 2], [0, 2], [-1, -1], [-1, -1], [-1, -1]],
            [[1, 1], [2, 2], [1, 2], [-1, -1], [-1, -1], [-1, -1]],
            # aa x bc -> ab or ac
            [[0, 0], [1, 2], [0, 1], [0, 2], [-1, -1], [-1, -1]],
            [[1, 1], [0, 2], [0, 1], [1, 2], [-1, -1], [-1, -1]],
            # ab x ab -> aa or ab or bb
            [[0, 1], [0, 1], [0, 0], [0, 1], [1, 1], [-1, -1]],
            [[1, 2], [1, 2], [1, 1], [1, 2], [2, 2], [-1, -1]],
            [[0, 2], [0, 2], [0, 0], [0, 2], [2, 2], [-1, -1]],
            # ab x bc -> ab or ac or bb or bc
            [[0, 1], [1, 2], [0, 1], [0, 2], [1, 1], [1, 2]],
            [[0, 1], [0, 2], [0, 0], [0, 1], [0, 1], [1, 2]],
            # ab x cd -> ac or ad or bc or bd
            [[0, 1], [2, 3], [0, 2], [0, 3], [1, 2], [1, 3]],
        ])
        expect = np.zeros((17, 4))
        self._test(genotypes, expect)

    def test_error_nonparental(self):
        genotypes = np.array([
            # aa x aa -> ab or ac or bb or cc
            [[0, 0], [0, 0], [0, 1], [0, 2], [1, 1], [2, 2]],
            [[1, 1], [1, 1], [0, 1], [1, 2], [0, 0], [2, 2]],
            [[2, 2], [2, 2], [0, 2], [1, 2], [0, 0], [1, 1]],
            # aa x ab -> ac or bc or cc
            [[0, 0], [0, 1], [0, 2], [1, 2], [2, 2], [2, 2]],
            [[0, 0], [0, 2], [0, 1], [1, 2], [1, 1], [1, 1]],
            [[1, 1], [0, 1], [1, 2], [0, 2], [2, 2], [2, 2]],
            # aa x bb -> ac or bc or cc
            [[0, 0], [1, 1], [0, 2], [1, 2], [2, 2], [2, 2]],
            [[0, 0], [2, 2], [0, 1], [1, 2], [1, 1], [1, 1]],
            [[1, 1], [2, 2], [0, 1], [0, 2], [0, 0], [0, 0]],
            # ab x ab -> ac or bc or cc
            [[0, 1], [0, 1], [0, 2], [1, 2], [2, 2], [2, 2]],
            [[0, 2], [0, 2], [0, 1], [1, 2], [1, 1], [1, 1]],
            [[1, 2], [1, 2], [0, 1], [0, 2], [0, 0], [0, 0]],
            # ab x bc -> ad or bd or cd or dd
            [[0, 1], [1, 2], [0, 3], [1, 3], [2, 3], [3, 3]],
            [[0, 1], [0, 2], [0, 3], [1, 3], [2, 3], [3, 3]],
            [[0, 2], [1, 2], [0, 3], [1, 3], [2, 3], [3, 3]],
            # ab x cd -> ae or be or ce or de
            [[0, 1], [2, 3], [0, 4], [1, 4], [2, 4], [3, 4]],
        ])
        expect = np.array([
            [1, 1, 2, 2],
            [1, 1, 2, 2],
            [1, 1, 2, 2],
            [1, 1, 2, 2],
            [1, 1, 2, 2],
            [1, 1, 2, 2],
            [1, 1, 2, 2],
            [1, 1, 2, 2],
            [1, 1, 2, 2],
            [1, 1, 2, 2],
            [1, 1, 2, 2],
            [1, 1, 2, 2],
            [1, 1, 1, 2],
            [1, 1, 1, 2],
            [1, 1, 1, 2],
            [1, 1, 1, 1],
        ])
        self._test(genotypes, expect)

    def test_error_hemiparental(self):
        genotypes = np.array([
            # aa x ab -> bb
            [[0, 0], [0, 1], [1, 1], [-1, -1]],
            [[0, 0], [0, 2], [2, 2], [-1, -1]],
            [[1, 1], [0, 1], [0, 0], [-1, -1]],
            # ab x bc -> aa or cc
            [[0, 1], [1, 2], [0, 0], [2, 2]],
            [[0, 1], [0, 2], [1, 1], [2, 2]],
            [[0, 2], [1, 2], [0, 0], [1, 1]],
            # ab x cd -> aa or bb or cc or dd
            [[0, 1], [2, 3], [0, 0], [1, 1]],
            [[0, 1], [2, 3], [2, 2], [3, 3]],
        ])
        expect = np.array([
            [1, 0],
            [1, 0],
            [1, 0],
            [1, 1],
            [1, 1],
            [1, 1],
            [1, 1],
            [1, 1],
        ])
        self._test(genotypes, expect)

    def test_error_uniparental(self):
        genotypes = np.array([
            # aa x bb -> aa or bb
            [[0, 0], [1, 1], [0, 0], [1, 1]],
            [[0, 0], [2, 2], [0, 0], [2, 2]],
            [[1, 1], [2, 2], [1, 1], [2, 2]],
            # ab x cd -> ab or cd
            [[0, 1], [2, 3], [0, 1], [2, 3]],
        ])
        expect = np.array([
            [1, 1],
            [1, 1],
            [1, 1],
            [1, 1],
        ])
        self._test(genotypes, expect)

    def test_parent_missing(self):
        genotypes = np.array([
            [[-1, -1], [0, 0], [0, 0], [1, 1]],
            [[0, 0], [-1, -1], [0, 0], [2, 2]],
            [[-1, -1], [-1, -1], [1, 1], [2, 2]],
        ])
        expect = np.array([
            [0, 0],
            [0, 0],
            [0, 0],
        ])
        self._test(genotypes, expect)



# -*- coding: utf-8 -*-
"""
Utility functions for running principal components analysis and plotting the
results.

See also the examples at:

- http://nbviewer.ipython.org/github/alimanfoo/anhima/blob/master/examples/pca.ipynb

"""  # noqa


from __future__ import division, print_function, absolute_import


# third party dependencies
import numpy as np
import matplotlib.pyplot as plt
import sklearn.decomposition


def pca(gn, n_components=10, whiten=False):
    """Perform a principal components analysis of genotypes, treating each
    variant as a feature.

    Parameters
    ----------

    gn : array_like, shape (`n_variants`, `n_samples`)
        A 2-dimensional array where each element is a genotype call coded as
        a single integer counting the number of non-reference alleles.
    n_components : int, None or string
        Number of components to keep. If `n_components` is None all
        components are kept: ``n_components == min(n_samples, n_features)``. If
        `n_components` == 'mle', Minka's MLE is used to guess the dimension. If
        0 < `n_components` < 1, select the number of components such that the
        amount of variance that needs to be explained is greater than the
        percentage specified by `n_components`.
    whiten : bool
        When True (False by default) the components vectors are divided by
        n_samples times singular values to ensure uncorrelated outputs with
        unit component-wise variances.

    Returns
    -------

    model : ``sklearn.decomposition.PCA``
        The fitted model.
    coords : ndarray, shape (`n_samples`, `n_components`)
        The result of fitting the model with `genotypes` and applying
        dimensionality reduction to `genotypes`.

    See Also
    --------

    sklearn.decomposition.PCA, anhima.ld.ld_prune_pairwise

    Notes
    -----

    The :func:`anhima.ld.ld_prune_pairwise` can be used to obtain a set of
    variants in approximate linkage equilibrium prior to running PCA.

    """

    # check inputs
    gn = np.asarray(gn)
    assert gn.ndim == 2

    # transpose because sklearn expects data as (n_samples, n_features)
    m = gn.T

    # set up PCA
    model = sklearn.decomposition.PCA(n_components=n_components, whiten=whiten,
                                      copy=True)

    # fit the model and apply dimensionality reduction
    coords = model.fit_transform(m)

    return model, coords


def plot_coords(model, coords, pcx=1, pcy=2, ax=None, colors='b', sizes=20,
                labels=None, scatter_kwargs=None, annotate_kwargs=None):
    """Scatter plot of transformed coordinates from principal components
    analysis.

    Parameters
    ----------

    model : ``sklearn.decomposition.PCA``
        The fitted model.
    coords : ndarray, shape (`n_samples`, `n_components`)
        The transformed coordinates.
    pcx : int, optional
        The principal component to plot on the X axis. N.B., this is
        one-based, so `1` is the first principal component, `2` is the second
        component, etc.
    pcy : int, optional
        The principal component to plot on the Y axis. N.B., this is
        one-based, so `1` is the first principal component, `2` is the second
        component, etc.
    ax : axes, optional
        The axes on which to draw. If not provided, a new figure will be
        created.
    colors : color or sequence of color, optional
        Can be a single color format string, or a sequence of color
        specifications of length `n_samples`.
    sizes : scalar or array_like, shape (`n_samples`), optional
        Size in points^2.
    labels : sequence of strings
        If provided, will be used to label points in the plot.
    scatter_kwargs : dict-like
        Additional keyword arguments passed through to `plt.scatter`.
    annotate_kwargs : dict-like
        Additional keyword arguments passed through to `plt.annotate` when
        labelling points.


    Returns
    -------

    ax : axes
        The axes on which the plot was drawn.

    """

    # check inputs
    coords = np.asarray(coords)
    assert coords.ndim == 2

    # set up axes
    if ax is None:
        # make a square figure
        x = plt.rcParams['figure.figsize'][0]
        fig, ax = plt.subplots(figsize=(x, x))

    # obtain X and Y data, N.B., `pcx` and `pcy` are 1-based
    x = coords[:, pcx-1]
    y = coords[:, pcy-1]

    # plot points
    if scatter_kwargs is None:
        scatter_kwargs = dict()
    ax.scatter(x, y, c=colors, s=sizes, **scatter_kwargs)

    # label points
    if labels is not None:
        if annotate_kwargs is None:
            annotate_kwargs = dict()
        annotate_kwargs.setdefault('xycoords', 'data')
        annotate_kwargs.setdefault('xytext', (3, 3))
        annotate_kwargs.setdefault('textcoords', 'offset points')
        for l, lx, ly in zip(labels, x, y):
            if l is not None:
                ax.annotate(str(l), xy=(lx, ly), **annotate_kwargs)

    # tidy up
    ax.set_xlabel('PC%s (%.2f%%)' %
                  (pcx, model.explained_variance_ratio_[pcx-1] * 100))
    ax.set_ylabel('PC%s (%.2f%%)' %
                  (pcy, model.explained_variance_ratio_[pcy-1] * 100))

    return ax


def plot_variance_explained(model, bar_kwargs=None, ax=None):
    """

    Parameters
    ----------

    model : ``sklearn.decomposition.PCA``
        The fitted model.
    bar_kwargs : dict-like, optional
        Additional keyword arguments passed through to ``ax.bar()``.
    ax : axes, optional
        The axes on which to draw. If not provided, a new figure will be
        created.

    Returns
    -------

    ax : axes
        The axes on which the plot was drawn.

    """

    # set up axes
    if ax is None:
        fig, ax = plt.subplots()

    # how many components are available

    # coordinates for bar
    y = model.explained_variance_ratio_ * 100  # express as percent
    n = len(y)
    x = np.arange(n)

    # plot bar
    if bar_kwargs is None:
        bar_kwargs = dict()
    bar_kwargs.setdefault('width', 1)
    ax.bar(x, y, **bar_kwargs)

    # tidy up
    ax.set_xticks(x+.5)
    ax.set_xticklabels(range(1, n+1))
    ax.set_xlabel('principal component')
    ax.set_ylabel('% variance explained')

    return ax


def plot_loadings(model, pc=1, pos=None, plot_kwargs=None, ax=None):
    """
    Plot loadings for the given principal component.

    Parameters
    ----------

    model : ``sklearn.decomposition.PCA``
        The fitted model.
    pc : int, optional
        The principal component to plot loadings for. N.B., this is
        one-based, so `1` is the first principal component, `2` is the second
        component, etc.
    pos : array_like, int, optional
        An array of variant positions to use for the X axis, If not given,
        variant index will be used for the X axis.
    plot_kwargs : dict-like, optional
        Additional keyword arguments passed through to ``ax.plot()``.
    ax : axes, optional
        The axes on which to draw. If not provided, a new figure will be
        created.

    Returns
    -------

    ax : axes
        The axes on which the plot was drawn.

    """

    # set up axes
    if ax is None:
        x = plt.rcParams['figure.figsize'][0]
        fig = plt.figure(figsize=(x, x//3))
        ax = fig.add_subplot(111)

    # obtain loadings
    y = model.components_[pc-1, :]

    # plot them
    if plot_kwargs is None:
        plot_kwargs = dict()
    if pos is not None:
        assert len(y) == len(pos)
        ax.plot(pos, y, **plot_kwargs)
        ax.set_xlabel('position')
        ax.set_xlim(min(pos), max(pos))
    else:
        ax.plot(y, **plot_kwargs)
        ax.set_xlabel('variant')
        ax.set_xlim(0, len(y))

    return ax
